@@ -5,13 +5,14 @@
 # (see pyrolab/__init__.py for details)
 
 """
-Lasers
-------
+Motion
+======
 
-Submodule containing drivers for each supported laser type.
+Submodule containing drivers for each supported motor type.
 """
 
 from pyrolab.drivers import Instrument
 
+
 class Motion(Instrument):
     pass
