@@ -5,13 +5,14 @@
 # (see pyrolab/__init__.py for details)
 
 """
-Lasers
-------
+Oscilloscopes
+=============
 
-Submodule containing drivers for each supported laser type.
+Submodule containing drivers for each supported scope type.
 """
 
 from pyrolab.drivers import Instrument
 
+
 class Scope(Instrument):
     pass
