@@ -6,7 +6,7 @@
 
 """
 Rohde & Schwarz Digital Oscilloscopes
--------------------------------------
+=====================================
 
 Submodule containing drivers for each supported laser type.
 
@@ -34,48 +34,107 @@ This manual describes the following R&S®RTO models with firmware version 3.70:
 
 If you don't have the NI VISA implementation installed on your computer, be 
 sure to install the separate dependency ``pyvisa-py``, which is not included
-with PyroLab.
+with PyroLab. NI VISA is available for Mac, Windows, and Linux.
+
+Common Issues
+-------------
+1. Note that if a trigger is set and you try to acquire data but end up with
+a timeout warning, it's possible that the acquisition never began because the
+trigger level was never reached. The scope will still be waiting to begin
+acquisition, but you'll be left without data and with a bad connection.
+
+.. admonition:: Dependencies
+   :class: note
+
+   | pyvisa
+   | NI-VISA *or* pyvisa-py
 """
 
+# Current Work
+# Check out manual, chapter 19.5.2.3
+# https://www.rohde-schwarz.com/us/applications/fast-remote-instrument-control-with-hislip-application-note_56280-30881.html
+# https://www.google.com/search?channel=tus5&client=firefox-b-1-d&q=pyvisa+hislip
+# https://github.com/pyvisa/pyvisa-py/issues/58
+
+# Even though VISAResourceExtensions is not used in this module, the act of 
+# importing it alone performs some monkey-patching on the pyvisa module, 
+# required by RTO. Don't remove this seemingly unused import!
+
 import time
 
-import deprecation
 import pyvisa as visa
 
 from pyrolab import __version__
 from pyrolab.drivers.scopes import Scope, VISAResourceExtentions
 
+
 class RTO(Scope):
     """
     Simple network controller class for R&S RTO oscilloscopes.
 
-    Parameters
-    ----------
-    address : str
-        The IP address of the instrument.
-    interface : str, optional
-        The interface to use to connect to the instrument. May be one
-        of "TCPIP", "GPIB", "ASRL", etc. Default is "TCPIP".
-    protocol : str, optional
-        The protocol to use for the LAN connection. Can be "INSTR"
-        or "hislip". Default is "hislip".
-    timeout : int, optional
-        The device response timeout in milliseconds (default infinite).
+    This class is used to control the R&S RTO oscilloscope. These are not local
+    devices, nor native PyroLab objects. Therefore, network device 
+    autodetection is not supported.
     """
-    def __init__(self, address, interface="TCPIP", protocol="hislip", timeout=None):
+    @staticmethod
+    def detect_devices():
+        """
+        Network device detection not supported.
+        
+        Becuase R&S oscilloscopes are connected to using the IP address,
+        this function does not detect them and instead always returns an empty 
+        list.
+        """
+        device_info = []
+        return device_info
+
+    def connect(self, address: str="", hislip: bool=False, timeout: float=1e3) -> bool:
+        """
+        Connects to and initializes the R&S RTO oscilloscope.
+        
+        HiSLIP (High-Speed LAN Instrument Protocol) is a TCP/IP-based protocol 
+        for remote instrument control of LAN-based test and measurement 
+        instruments. It is intended to replace the older VXI-11 protocol.
+
+        .. warning::
+           The HiSLIP protocol is not supported when using the pyvisa-py
+           backend **on the client machine**. To use it, you must use the NI VISA 
+           implementation instead.
+
+        Parameters
+        ----------
+        address : str
+            The IP address of the instrument.
+        hislip : bool, optional
+            Whether to use the HiSLIP protocol or not (default ``False``).
+        timeout : int, optional
+            The device response timeout in milliseconds (default 1 ms).
+            Pass ``None`` for infinite timeout.
+        """
         rm = visa.ResourceManager()
-        self.device = rm.open_resource("{}::{}::{}".format(interface, address, protocol))
+        if hislip:
+            self.device = rm.open_resource(f"TCPIP::{address}::hislip0")    
+        else:
+            self.device = rm.open_resource(f"TCPIP::{address}")
         self.device.timeout = timeout
         self.write_termination = ''
         self.device.ext_clear_status()
         
-        # print("Connected: {}".format(self.device.query('*IDN?')))
         self.write('*RST;*CLS')
         self.write('SYST:DISP:UPD ON')
         self.device.ext_error_checking()
 
+        return True
+
+    def close(self):
+        self.device.close()
+
     @property
     def timeout(self):
+        """
+        Network timeout duration in milliseconds (errors out if no response
+        received within timeout).
+        """
         return self.device.timeout
 
     @timeout.setter
@@ -84,7 +143,7 @@ class RTO(Scope):
 
     def query(self, message, delay=None):
         """
-        A combination of :py:func:`write(message)` and :py:func:`read()`.
+        A combination of :py:func:`write` and :py:func:`read`.
 
         Parameters
         ----------
@@ -92,9 +151,9 @@ class RTO(Scope):
             The message to send.
         delay : float, optional
             Delay in seconds between write and read operations. If None,
-            defaults to `self.device.query_delay`.
+            defaults to ``self.device.query_delay``.
         """
-        self.device.query(message, delay)
+        return self.device.query(message, delay)
 
     def write(self, message, termination=None, encoding=None):
         """
@@ -130,41 +189,14 @@ class RTO(Scope):
         self.wait_for_device()
         self.device.ext_error_checking()
 
-    @deprecation.deprecated(deprecated_in="0.1.0", removed_in="0.2.0",
-                current_version=__version__,
-                details="Use 'write_block()' instead.")
-    def __send_command(self, command):
-        """
-        Writes a message to the scope, waits for it to complete, and checks for errors.
-
-        Warning
-        -------
-        .. deprecated:: 0.1.0
-           :py:func:`__send_command` will be removed in 0.2.0, it is replaced by
-           :py:func:`write_block` beginning in 0.1.0.
-
-        Parameters
-        ----------
-        command : str
-            The message to send.
-
-        Notes
-        -----
-        This function is blocking.
-        """
-        self.device.write(command)
-        self.wait_for_device()
-        self.device.ext_error_checking()
-
     def wait_for_device(self):
         """
         Waits for the device until last action is complete.
 
-        Notes
-        -----
-        This function is blocking.
+        .. note::
+           This function is blocking.
         """
-        self.device.query('*OPC?')
+        self.device.ext_wait_for_opc()
 
     def acquisition_settings(self, sample_rate, duration, force_realtime=False):
         """
@@ -173,21 +205,37 @@ class RTO(Scope):
         The exact command this executes is:
         `'ACQ:POIN:AUTO RES;:ACQ:SRAT {};:TIM:RANG {}'`
 
+        .. warning::
+
+            The oscilloscope has a record length limit. This is system 
+            dependent! If you are getting cryptic errors, such as "Data out of 
+            range;ACQ:SRAT <some value>", you might be exceeding the record
+            length limit, calculated as SAMPLE_RATE x DURATION. The system 
+            enforces this limit to "prevent undersampling and ensure a 
+            sufficient resolution to acquire the correct waveform if the time 
+            scale is changed." Check your specific system to find the record
+            length limit.
+
+            See also: RTO User Manual Chapter 4.2 (page 147).
+
         Parameters
         ----------
         sample_rate : float
-            Sample rate of device.
+            Sample rate of device in Sa/s. Range is 2 to 20e+12 in increments
+            of 1.
         duration : float
             Length of acquisition in seconds.
         force_realtime : bool, optional
             Defaults to False.
         """
-        short_command = 'ACQ:POIN:AUTO RES;:TIM:RANG {}'.format(duration)
+        short_command = 'ACQ:POIN:AUTO RES;:TIM:RANG {};:ACQ:SRAT {}'.format(duration, sample_rate)
         if force_realtime:
             self.write_block('ACQ:MODE RTIM')
         self.write_block(short_command)
 
-    def set_channel(self, channel, state="ON", coupling="DCLimit", range=0.5, position=0, offset=0, invert="OFF"):
+    def set_channel(self, channel: int, state: str="ON", 
+                    coupling: str="DCLimit", range: float=0.5, 
+                    position: float=0, offset: float=0, invert: str="OFF"):
         """
         Configures a channel.
 
@@ -202,7 +250,7 @@ class RTO(Scope):
             Selects the connection of the indicated channel signal. Valid values are
             "DC" (direct connection with 50 ohm termination), "DCLimit" (direct connection 
             with 1M ohm termination), or "AC" (connection through DC capacitor).
-            Default is DCLimit (see ``CHANnel<m>:COUPling``).
+            Default is "DCLimit" (see ``CHANnel<m>:COUPling``).
         range : float, optional
             Sets the voltage range across the 10 vertical divisions of the diagram
             in V/div. Default is 0.5 (see ``CHANnel<m>:RANGe``).
@@ -226,23 +274,6 @@ class RTO(Scope):
         )
         self.write_block(cmd)
 
-    @deprecation.deprecated(deprecated_in="0.1.0", removed_in="0.2.0",
-                current_version=__version__,
-                details="Use 'set_channel()' instead.")
-    def add_channel(self, channel_num, range, position = 0, offset = 0, coupling = "DCL"):
-        """Add a channel.
-        
-        Warning
-        -------
-        .. deprecated:: 0.1.0
-           :py:func:`add_channel` will be removed in 0.2.0, it is replaced by
-           :py:func:`set_channel` beginning in 0.1.0.
-        """
-        short_command = 'CHAN{}:RANG {};POS {};OFFS {};COUP {};STAT ON'.format(
-            channel_num, range, position, offset, coupling
-        )
-        self.__send_command(short_command)
-
     def __add_trigger(self,
         type,
         source,
@@ -250,15 +281,13 @@ class RTO(Scope):
         level,
         trigger_num = 1,
         mode = "NORM",
-        settings: str = ""
-    ):
+        settings: str = ""):
         short_command = 'TRIG{}:MODE {};SOUR {};TYPE {};LEV{} {};'.format(
-        trigger_num, mode, source, type, source_num, level
-    )
+        trigger_num, mode, source, type, source_num, level)
         #Add a trigger.
         self.write_block(short_command + settings)
 
-    def edge_trigger(self, channel, voltage):
+    def edge_trigger(self, channel: int, voltage: float) -> None:
         """
         Add an edge trigger to the specified channel.
 
@@ -278,18 +307,30 @@ class RTO(Scope):
             settings = "EDGE:SLOP POS"
         )
 
-    def acquire(self, run='single'):
+    def acquire(self, timeout: int=-1, run: str='single') -> None:
         """
         Asynchronous command that starts acquisition.
 
         Parameters
         ----------
+        timeout : int
+            The timeout to use for the given acquisition in milliseconds. The 
+            default timeout is used if no timeout is given. The timeout is not
+            permanently set; once the acquisition is complete, the original 
+            timeout is kept for any other query/write command. For an infinite
+            timeout, pass in None. If not specified, default timeout is the 
+            default for `acquire()`.
         run : str
             Specifies the type of run. Allowable values are ``continuous`` 
             (starts the continuous acquisition), ``single`` (starts a defined
             number of acquisition cycles as set by 
-            :py:func:``acquisition_settings()``), or ``stop`` (stops a 
+            :py:func:`acquisition_settings`), or ``stop`` (stops a 
             running acquisition). Default is ``single``.
+
+        Raises
+        ------
+        ValueError
+            If the run type is not one of the allowed values.
         """        
         if run == "single":
             cmd = "SING"
@@ -299,43 +340,83 @@ class RTO(Scope):
             cmd = "STOP"
         else:
             raise ValueError("%s is not a valid argument" % run)
+
+        if timeout != -1:
+            default_timeout = self.timeout
+            self.timeout = timeout
         
         self.write(cmd)
 
-    @deprecation.deprecated(deprecated_in="0.1.0", removed_in="0.2.0",
-                current_version=__version__,
-                details="Use 'acquire()' instead.")
-    def start_acquisition(self, timeout, type='SING'):
-        """
-        Asynchronous command that starts acquisition.
+        if timeout != -1:
+            self.timeout = default_timeout
 
-        Warning
-        -------
-        .. deprecated:: 0.1.0
-           :py:func:`start_acquisition` will be removed in 0.2.0, it is replaced by
-           :py:func:`acquire` beginning in 0.1.0.
+    def set_timescale(self, time: float) -> None:
+        """
+        Sets the horizontal scale--the time per division on the x-axis--for all 
+        channel and math waveforms.
 
         Parameters
         ----------
-        timeout : int
-            The timeout in seconds for all I/O operations.
-        run : str
-            Specifies the type of run. Allowable values are ``continuous`` 
-            (starts the continuous acquisition), ``single`` (starts a defined
-            number of acquisition cycles as set by ``acquisition_settings()``),
-            or ``stop`` (stops a running acquisition). Default is ``single``.
-        """        
-        # Translate seconds to ms.
-        self.device.timeout = timeout * 1000
-        if type not in ["SING", "RUN", "STOP"]:
-            raise ValueError("%s is not a valid argument" % type)            
+        time : float
+            The time (in seconds) per division. Valid range is from 25e-12 to 
+            10000 (RTO) | 5000 (RTE) in increments of 1e-12. (`*RST` value is 
+            10e-9).
+        """
+        self.write(f'TIM:SCAL {str(time)}')
+        self.device.ext_error_checking()
+
+    def set_auto_measurement(self, measurement: int=1, source: str='C1W1', 
+                             meastype: str='MAX') -> None:
+        """
+        Convenience function for setting default measurements. 
         
-        self.write(type)
+        Use if you haven't configured any of your own measurements.
+
+        Parameters
+        ----------
+        measurement : int
+            The oscope supports storing up to 8 measurements. Default is 1.
+        source : str
+            The source to setup the measurement on. See page 1377 of the User
+            Manual for valid sources. Common ones are of the format "C<m>W<n>",
+            where <m> is the channel and <n> is the waveform (e.g., "C1W1", 
+            which is the default).
+        meastype : str
+            The measurement type to associate with the measurement. See page 
+            1381 of the User Manual for valid measurement types. Common ones
+            are 'HIGH', 'LOW', 'AMPLitude', 'MAXimum', 'MINimum', 'MEAN'. 
+            Default is 'MAX'.
+        """
+        self.write(f'MEAS{measurement}:SOUR {source}')
+        # "MAX" will measure the max value in the current view window (Based on time base)
+        self.write(f'MEAS{measurement}:MAIN {meastype}')
+        self.write(f'MEAS{measurement} ON')
+
+    def measure(self, measurement: int=1) -> float:
+        """
+        Takes the result of an automatic measurement as setup using 
+        `set_auto_measurement()`.
+
+        Parameters
+        ----------
+        measurement : int
+            The measurement to take. Default is 1.
+
+        Returns
+        -------
+        float
+            The result of the automatic measurement.
+        """
+        return float(self.query(f'MEAS{measurement}:RES:ACT?'))
 
     def get_data(self, channel, form="ascii"):
         """
-        Retrieves waveform data from the specified channel in the specified 
-        data type.
+        Retrieves waveform data from the specified channel
+        
+        Data is retrieved in the specified 
+        data type. Note that data like this can be transferred in two ways: 
+        in ASCII form (slow, but human readable) and binary (fast, but more 
+        difficult to debug).
 
         Parameters
         ----------
@@ -346,8 +427,8 @@ class RTO(Scope):
             Allowable values are ``ascii``, ``real``, ``int8``, and ``int16``.
             Default is ``ascii``.
 
-        See Also
-        --------
+        Notes
+        -----
         RTO User Manual, commands for ``FORMat[:DATA]`` and 
         ``CHANnel<m>[:WAVeform<n>]:DATA[:VALues]?``
         """
@@ -364,47 +445,16 @@ class RTO(Scope):
 
         cmd = 'FORM {};:CHAN{}:DATA?'.format(fmt, channel)
         
+        # Consider skipping the intermediate "list" step and having pyvisa
+        # automatically convert to a numpy array (see
+        # https://pyvisa.readthedocs.io/en/latest/introduction/rvalues.html#reading-ascii-values)
         if form == "ascii":
             return self.device.query_ascii_values(cmd)
         elif form == "real":
             return self.device.query_binary_values(cmd)
         else:
             return self.query(cmd)
-
-    @deprecation.deprecated(deprecated_in="0.1.0", removed_in="0.2.0",
-                current_version=__version__,
-                details="Use 'get_data()' instead.")
-    def get_data_ascii(self, channel):
-        """
-        Get the data in ascii encoding.
-
-        Warning
-        -------
-        .. deprecated:: 0.1.0
-           :py:func:`get_data_ascii` will be removed in 0.2.0, it is replaced by
-           :py:func:`get_data` beginning in 0.1.0.
-        """
-        dataQuery = 'FORM ASC;:CHAN{}:DATA?'.format(channel)
-        waveform = self.device.query_ascii_values(dataQuery)
-        return waveform
-
-    @deprecation.deprecated(deprecated_in="0.1.0", removed_in="0.2.0",
-                current_version=__version__,
-                details="Use 'get_data()' instead.")
-    def get_data_binary(self, channel):
-        """
-        Get the data in binary encoding.
-
-        Warning
-        -------
-        .. deprecated:: 0.1.0
-           :py:func:`get_data_binary` will be removed in 0.2.0, it is replaced by
-           :py:func:`get_data` beginning in 0.1.0.
-        """
-        dataQuery = 'FORM REAL;:CHAN{}:DATA?'.format(channel)
-        waveform = self.device.query_binary_values(dataQuery)
-        return waveform
-
+    
     def screenshot(self, path):
         """
         Takes a screenshot of the scope and saves it to the specified path.
@@ -429,38 +479,53 @@ class RTO(Scope):
         )
         self.device.ext_error_checking()
 
-    @deprecation.deprecated(deprecated_in="0.1.0", removed_in="0.2.0",
-                current_version=__version__,
-                details="Use 'screenshot()' instead.")
-    def take_screenshot(self, path):
+    def act_filter(self, channel):
         """
-        Takes a screenshot of the scope and saves it to the specified path.
-
-        Image format is PNG.
-
-        Warning
-        -------
-        .. deprecated:: 0.1.0
-           :py:func:`take_screenshot` will be removed in 0.2.0, it is replaced by
-           :py:func:`screenshot` beginning in 0.1.0.
+        Activates the lowpass filter for a channel. 
 
         Parameters
         ----------
-        path : str
-            The local path, including filename and extension, of where
-            to save the file.
+        channel : int
+            The channel (1-4) on which to activate the filter.
         """
-        instrument_save_path = '\'C:\\temp\\Last_Screenshot.png\''
-        self.device.write('HCOP:DEV:LANG PNG')
-        self.device.write('MMEM:NAME {}'.format(instrument_save_path))
-        self.device.write('HCOP:IMM')
-        self.wait_for_device()
-        self.device.ext_error_checking()
-        self.device.ext_query_bin_data_to_file(
-            'MMEM:DATA? {}'.format(instrument_save_path),
-            str(path)
-        )
-        self.device.ext_error_checking()
+        self.write(f"CHAN{channel}:DIGF:STAT ON")
+
+    def deact_filter(self, channel):
+        """
+        Deactivates the lowpass filter on a given channel.
+
+        Parameters
+        ----------
+        channel : int
+            The channel (1-4) where the new cutoff frequency is applied.
+        """
+        self.write(f"CHAN{channel}:DIGF:STAT OFF")
+    
+    def set_cutoff_freq(self, channel, cutoff_freq):
+        """
+        Sets the cutoff frequency of one of the two filters: either the filter 
+        for channels 1 and 2 or the filter for channels 3 and 4. 
+        Cutoff frequencies are set only for either channels 1 and 2 or 
+        channels 3 and 4, but you can activate the filter for each channel 
+        separately.
+
+        Parameters
+        ----------
+        channel : int
+            Specifies which filter's cutoff frequency will be changed. A value 
+            of 1 or 2 will set the cutoff frequency for both channels, and a 
+            value of 3 or 4 will do the same for both channels 3 and 4.
+        cutoff_freq : int
+            The cutoff frequency enabled on the channel. Must be between 100 
+            kHz and 1 GHz or 2 GHz (depending on the scope). The scope only supports certain discrete cutoff 
+            frequencies. Any other frequency will be rounded to the closest 
+            frequency the scope supports.
+        """
+        if cutoff_freq >= 1e5 and cutoff_freq <= 2e9:
+            self.write(f"CHAN{channel}:DIGF:CUT {cutoff_freq}")
+        else:
+            raise ValueError("Cutoff frequency must be between 1e5 and 2e9 Hz")
+
 
 class RemoteDisplay:
     def __init__(self, scope: RTO):
